@@ -1,29 +1,39 @@
+from __future__ import absolute_import
+from __future__ import division
+from __future__ import print_function
+from __future__ import unicode_literals
+
+import six
+import tty, sys
 import curses, datetime, locale
 from decimal import Decimal
-_ = lambda x:x
-from electrum.util import format_satoshis, set_verbosity
-from electrum.util import StoreDict
-from electrum.bitcoin import is_valid, COIN
+import getpass
 
+import electrum
+from electrum.util import format_satoshis, set_verbosity
+from electrum.bitcoin import is_address, COIN, TYPE_ADDRESS
 from electrum import Wallet, WalletStorage
 
-import tty, sys
+_ = lambda x:x
+
 
 
 class ElectrumGui:
 
-    def __init__(self, config, network, plugins):
+    def __init__(self, config, daemon, plugins):
 
         self.config = config
-        self.network = network
+        self.network = daemon.network
         storage = WalletStorage(config.get_wallet_path())
         if not storage.file_exists:
-            print "Wallet not found. try 'electrum create'"
+            print("Wallet not found. try 'electrum create'")
             exit()
-
+        if storage.is_encrypted():
+            password = getpass.getpass('Password:', stream=None)
+            storage.decrypt(password)
         self.wallet = Wallet(storage)
         self.wallet.start_threads(self.network)
-        self.contacts = StoreDict(self.config, 'contacts')
+        self.contacts = self.wallet.contacts
 
         locale.setlocale(locale.LC_ALL, '')
         self.encoding = locale.getpreferredencoding()
@@ -35,6 +45,7 @@ class ElectrumGui:
         curses.use_default_colors()
         curses.init_pair(1, curses.COLOR_WHITE, curses.COLOR_BLUE)
         curses.init_pair(2, curses.COLOR_WHITE, curses.COLOR_CYAN)
+        curses.init_pair(3, curses.COLOR_BLACK, curses.COLOR_WHITE)
         self.stdscr.keypad(1)
         self.stdscr.border(0)
         self.maxy, self.maxx = self.stdscr.getmaxyx()
@@ -53,7 +64,7 @@ class ElectrumGui:
         self.history = None
 
         if self.network:
-            self.network.register_callback('updated', self.update)
+            self.network.register_callback(self.update, ['updated'])
 
         self.tab_names = [_("History"), _("Send"), _("Receive"), _("Addresses"), _("Contacts"), _("Banner")]
         self.num_tabs = len(self.tab_names)
@@ -80,7 +91,7 @@ class ElectrumGui:
         self.set_cursor(0)
         return s
 
-    def update(self):
+    def update(self, event):
         self.update_history()
         if self.tab == 0:
             self.print_history()
@@ -104,18 +115,17 @@ class ElectrumGui:
 
         b = 0
         self.history = []
-
         for item in self.wallet.get_history():
-            tx_hash, conf, value, timestamp, balance = item
+            tx_hash, height, conf, timestamp, value, balance = item
             if conf:
                 try:
                     time_str = datetime.datetime.fromtimestamp(timestamp).isoformat(' ')[:-3]
                 except Exception:
                     time_str = "------"
             else:
-                time_str = 'pending'
+                time_str = 'unconfirmed'
 
-            label, is_default_label = self.wallet.get_label(tx_hash)
+            label = self.wallet.get_label(tx_hash)
             if len(label) > 40:
                 label = label[0:37] + '...'
             self.history.append( format_str%( time_str, label, format_satoshis(value, whitespaces=True), format_satoshis(balance, whitespaces=True) ) )
@@ -145,7 +155,7 @@ class ElectrumGui:
         self.stdscr.addstr(self.maxy -1, self.maxx-30, ' '.join([_("Settings"), _("Network"), _("Quit")]))
 
     def print_receive(self):
-        addr = self.wallet.get_unused_address(None)
+        addr = self.wallet.get_receiving_address()
         self.stdscr.addstr(2, 1, "Address: "+addr)
         self.print_qr(addr)
 
@@ -155,7 +165,7 @@ class ElectrumGui:
 
     def print_addresses(self):
         fmt = "%-35s  %-30s"
-        messages = map(lambda addr: fmt % (addr, self.wallet.labels.get(addr,"")), self.wallet.addresses())
+        messages = map(lambda addr: fmt % (addr, self.wallet.labels.get(addr,"")), self.wallet.get_addresses())
         self.print_list(messages,   fmt % ("Address", "Label"))
 
     def print_edit_line(self, y, label, text, index, size):
@@ -171,31 +181,38 @@ class ElectrumGui:
         self.print_edit_line(9, _("Fee"), self.str_fee, 3, 15)
         self.stdscr.addstr( 12, 15, _("[Send]"), curses.A_REVERSE if self.pos%6==4 else curses.color_pair(2))
         self.stdscr.addstr( 12, 25, _("[Clear]"), curses.A_REVERSE if self.pos%6==5 else curses.color_pair(2))
+        self.maxpos = 6
 
     def print_banner(self):
         if self.network:
             self.print_list( self.network.banner.split('\n'))
 
     def print_qr(self, data):
-        import qrcode, StringIO
-        s = StringIO.StringIO()
+        import qrcode
+        try:
+            from StringIO import StringIO
+        except ImportError:
+            from io import StringIO
+
+        s = StringIO()
         self.qr = qrcode.QRCode()
         self.qr.add_data(data)
-        self.qr.print_ascii(out=s, invert=True)
+        self.qr.print_ascii(out=s, invert=False)
         msg = s.getvalue()
         lines = msg.split('\n')
         for i, l in enumerate(lines):
             l = l.encode("utf-8")
-            self.stdscr.addstr(i+5, 5, l)
+            self.stdscr.addstr(i+5, 5, l, curses.color_pair(3))
 
-    def print_list(self, list, firstline = None):
-        self.maxpos = len(list)
+    def print_list(self, lst, firstline = None):
+        lst = list(lst)
+        self.maxpos = len(lst)
         if not self.maxpos: return
         if firstline:
             firstline += " "*(self.maxx -2 - len(firstline))
             self.stdscr.addstr( 1, 1, firstline )
         for i in range(self.maxy-4):
-            msg = list[i] if i < len(list) else ""
+            msg = lst[i] if i < len(lst) else ""
             msg += " "*(self.maxx - 2 - len(msg))
             m = msg[0:self.maxx - 2]
             m = m.encode(self.encoding)
@@ -209,15 +226,16 @@ class ElectrumGui:
 
     def main_command(self):
         c = self.stdscr.getch()
-        print c
+        print(c)
+        cc = curses.unctrl(c).decode()
         if   c == curses.KEY_RIGHT: self.tab = (self.tab + 1)%self.num_tabs
         elif c == curses.KEY_LEFT: self.tab = (self.tab - 1)%self.num_tabs
         elif c == curses.KEY_DOWN: self.pos +=1
         elif c == curses.KEY_UP: self.pos -= 1
         elif c == 9: self.pos +=1 # tab
-        elif curses.unctrl(c) in ['^W', '^C', '^X', '^Q']: self.tab = -1
-        elif curses.unctrl(c) in ['^N']: self.network_dialog()
-        elif curses.unctrl(c) == '^S': self.settings_dialog()
+        elif cc in ['^W', '^C', '^X', '^Q']: self.tab = -1
+        elif cc in ['^N']: self.network_dialog()
+        elif cc == '^S': self.settings_dialog()
         else: return c
         if self.pos<0: self.pos=0
         if self.pos>=self.maxpos: self.pos=self.maxpos - 1
@@ -238,10 +256,11 @@ class ElectrumGui:
 
     def edit_str(self, target, c, is_num=False):
         # detect backspace
+        cc = curses.unctrl(c).decode()
         if c in [8, 127, 263] and target:
             target = target[:-1]
-        elif not is_num or curses.unctrl(c) in '0123456789.':
-            target += curses.unctrl(c)
+        elif not is_num or cc in '0123456789.':
+            target += cc
         return target
 
 
@@ -266,8 +285,8 @@ class ElectrumGui:
 
     def run_contacts_tab(self, c):
         if c == 10 and self.contacts:
-            out = self.run_popup('Adress', ["Copy", "Pay to", "Edit label", "Delete"]).get('button')
-            key = self.contacts.keys()[self.pos%len(self.contacts.keys())]
+            out = self.run_popup('Address', ["Copy", "Pay to", "Edit label", "Delete"]).get('button')
+            key = list(self.contacts.keys())[self.pos%len(self.contacts.keys())]
             if out == "Pay to":
                 self.tab = 1
                 self.str_recipient = key
@@ -275,7 +294,7 @@ class ElectrumGui:
             elif out == "Edit label":
                 s = self.get_string(6 + self.pos, 18)
                 if s:
-                    self.wallet.labels[address] = s
+                    self.wallet.labels[key] = s
 
     def run_banner_tab(self, c):
         self.show_message(repr(c))
@@ -306,7 +325,7 @@ class ElectrumGui:
         self.str_description = ''
 
     def do_send(self):
-        if not is_valid(self.str_recipient):
+        if not is_address(self.str_recipient):
             self.show_message(_('Invalid Bitcoin address'))
             return
         try:
@@ -320,15 +339,14 @@ class ElectrumGui:
             self.show_message(_('Invalid Fee'))
             return
 
-        if self.wallet.use_encryption:
+        if self.wallet.has_password():
             password = self.password_dialog()
             if not password:
                 return
         else:
             password = None
-
         try:
-            tx = self.wallet.mktx( [("address", self.str_recipient, amount)], password, self.config, fee)
+            tx = self.wallet.mktx([(TYPE_ADDRESS, self.str_recipient, amount)], password, self.config, fee)
         except Exception as e:
             self.show_message(str(e))
             return
@@ -336,10 +354,8 @@ class ElectrumGui:
         if self.str_description:
             self.wallet.labels[tx.hash()] = self.str_description
 
-        h = self.wallet.send_tx(tx)
         self.show_message(_("Please wait..."), getchar=False)
-        self.wallet.tx_event.wait()
-        status, msg = self.wallet.receive_tx( h, tx )
+        status, msg = self.network.broadcast(tx)
 
         if status:
             self.show_message(_('Payment sent.'))
@@ -358,16 +374,15 @@ class ElectrumGui:
         w.refresh()
         if getchar: c = self.stdscr.getch()
 
-
     def run_popup(self, title, items):
-        return self.run_dialog(title, map(lambda x: {'type':'button','label':x}, items), interval=1, y_pos = self.pos+3)
-
+        return self.run_dialog(title, list(map(lambda x: {'type':'button','label':x}, items)), interval=1, y_pos = self.pos+3)
 
     def network_dialog(self):
-        if not self.network: return
-        host, port, protocol, proxy_config, auto_connect = self.network.get_parameters()
+        if not self.network:
+            return
+        params = self.network.get_parameters()
+        host, port, protocol, proxy_config, auto_connect = params
         srv = 'auto-connect' if auto_connect else self.network.default_server
-
         out = self.run_dialog('Network', [
             {'label':'server', 'type':'str', 'value':srv},
             {'label':'proxy', 'type':'str', 'value':self.config.get('proxy', '')},
@@ -382,24 +397,16 @@ class ElectrumGui:
                     except Exception:
                         self.show_message("Error:" + server + "\nIn doubt, type \"auto-connect\"")
                         return False
-
-                if out.get('proxy'):
-                    proxy = self.parse_proxy_options(out.get('proxy'))
-                else:
-                    proxy = None
-
+            if out.get('server') or out.get('proxy'):
+                proxy = electrum.network.deserialize_proxy(out.get('proxy')) if out.get('proxy') else proxy_config
                 self.network.set_parameters(host, port, protocol, proxy, auto_connect)
 
-
-
     def settings_dialog(self):
+        fee = str(Decimal(self.config.fee_per_kb()) / COIN)
         out = self.run_dialog('Settings', [
-            {'label':'Default GUI', 'type':'list', 'choices':['classic','lite','gtk','text'], 'value':self.config.get('gui')},
-            {'label':'Default fee', 'type':'satoshis', 'value': format_satoshis(self.wallet.fee_per_kb).strip() }
+            {'label':'Default fee', 'type':'satoshis', 'value': fee }
             ], buttons = 1)
         if out:
-            if out.get('Default GUI'):
-                self.config.set_key('gui', out['Default GUI'], True)
             if out.get('Default fee'):
                 fee = int(Decimal(out['Default fee']) * COIN)
                 self.config.set_key('fee_per_kb', fee, True)
@@ -415,7 +422,7 @@ class ElectrumGui:
     def run_dialog(self, title, items, interval=2, buttons=None, y_pos=3):
         self.popup_pos = 0
 
-        self.w = curses.newwin( 5 + len(items)*interval + (2 if buttons else 0), 50, y_pos, 5)
+        self.w = curses.newwin( 5 + len(list(items))*interval + (2 if buttons else 0), 50, y_pos, 5)
         w = self.w
         out = {}
         while True:
@@ -423,7 +430,7 @@ class ElectrumGui:
             w.border(0)
             w.addstr( 0, 2, title)
 
-            num = len(items)
+            num = len(list(items))
 
             numpos = num
             if buttons: numpos += 2
@@ -446,14 +453,14 @@ class ElectrumGui:
                 if len(value)<20:
                     value += ' '*(20-len(value))
 
-                if item.has_key('value'):
+                if 'value' in item:
                     w.addstr( 2+interval*i, 2, label)
                     w.addstr( 2+interval*i, 15, value, curses.A_REVERSE if self.popup_pos%numpos==i else curses.color_pair(1) )
                 else:
                     w.addstr( 2+interval*i, 2, label, curses.A_REVERSE if self.popup_pos%numpos==i else 0)
 
             if buttons:
-                w.addstr( 5+interval*i, 10, "[  ok  ]",     curses.A_REVERSE if self.popup_pos%numpos==(numpos-2) else curses.color_pair(2))
+                w.addstr( 5+interval*i, 10, "[  ok  ]", curses.A_REVERSE if self.popup_pos%numpos==(numpos-2) else curses.color_pair(2))
                 w.addstr( 5+interval*i, 25, "[cancel]", curses.A_REVERSE if self.popup_pos%numpos==(numpos-1) else curses.color_pair(2))
 
             w.refresh()
